import os

import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..image.load import get_shape, load_data
from ..image.mask import get_mask, get_mask_range
from ..image.std import calc_std
from ..util.dataset import normalized, masked
from ..util.tfrecord import save_tfrecord
from ..util.pickle import save_pickle
from ..util.numpy import save_numpy


class DataCommand(Command):

    name = 'data'
    description = 'Create TFRecord'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return False

    def is_target(self, stage):
        return stage == 0

    def force_stage(self, stage):
        return 0

    def create(self, data, prev, curr, log, imgs_file, mask_type):
        def _gen():
            for x in imgs:
                yield tf.convert_to_tensor(wrap(x)[y0:y1, x0:x1], tf.float32)

        imgs_file = os.path.join(self.application.job_dir, imgs_file)
        batch = self.status.params['batch']
        verbose = self.status.params['pbar']

        imgs, wrap = load_data(imgs_file)
        nt, h, w = get_shape(imgs_file)
        mask = get_mask(mask_type, h, w, self.application.job_dir)

        y0, y1, x0, x1 = get_mask_range(mask)
        data = tf.data.Dataset.from_generator(_gen, tf.float32)
        mask = mask[y0:y1, x0:x1]

        avgt, avgx, std = calc_std(data.batch(batch), mask, nt, verbose)
        nx = np.count_nonzero(mask)
        stat = nx, nt, h, w, y0, x0, std

        normalized_data = normalized(data, avgt, avgx, std)
        masked_data = masked(normalized_data, mask)

        save_tfrecord(f'{curr}-data', masked_data, nt, verbose)
        save_numpy(f'{curr}-mask', mask)
        save_numpy(f'{curr}-avgt', avgt)
        save_numpy(f'{curr}-avgx', avgx)
        save_pickle(f'{curr}-stat', stat)



import tensorflow.keras.backend as K
import tensorflow as tf

from ..util.distribute import distributed, ReduceOp
from ..util.dataset import unmasked
from ..image.filter.gaussian import gaussian
from ..image.filter.laplace import gaussian_laplace_multi
from .segment import get_segment_index
from .util import get_normalized_val, get_magnitude, ToDense


def make_segment(dataset, mask, gauss, radius, peaks, shard, batch, verbose):
    strategy = tf.distribute.get_strategy()

    ts, rs, ys, xs = peaks[:, 0], peaks[:, 1], peaks[:, 2], peaks[:, 3]
    nk = ts.size

    dataset = unmasked(dataset.shard(shard, 0), mask)
    dataset = dataset.batch(batch)
    to_dense = ToDense(mask)

    gauss = K.constant(gauss, tf.float32)
    radius = K.constant(radius, tf.float32)
    ts = K.constant(ts / shard, tf.int32)
    rs = K.constant(rs, tf.int32)
    ys = K.constant(ys, tf.int32)
    xs = K.constant(xs, tf.int32)

    prog = tf.keras.utils.Progbar(nk, verbose=verbose)
    fps = tf.TensorArray(tf.float32, 0, True)
    e = K.constant(0, tf.int32)
    for imgs in strategy.experimental_distribute_dataset(dataset):
        s, e = e, e + tf.shape(imgs)[0]
        gl, yl, xl = _prepare(
            imgs, s, e, ts, rs, ys, xs, gauss, radius,
        )
        nk = tf.size(yl)
        for k in tf.range(nk):
            g, y, x = gl[k], yl[k], xl[k]
            pos = get_segment_index(g, y, x, mask)
            val = get_normalized_val(g, pos)
            if get_magnitude(g, pos) > 0.0:
                footprint = to_dense(pos, val)
                fps = fps.write(fps.size(), footprint)
            prog.add(1)
    return fps.stack().numpy()


@distributed(*[ReduceOp.CONCAT for _ in range(5)], loop=False)
def _prepare(imgs, start, end, ts, rs, ys, xs, gauss, radius):
    cond = (start <= ts) & (ts < end)
    ids = tf.cast(tf.where(cond)[:, 0], tf.int32)
    tl = tf.gather(ts, ids) - start
    rl = tf.gather(rs, ids)
    yl = tf.gather(ys, ids)
    xl = tf.gather(xs, ids)
    if gauss > 0.0:
        imgs = gaussian(imgs, gauss)
    gls = gaussian_laplace_multi(imgs, radius)
    gl = tf.gather_nd(gls, tf.stack([tl, rl], 1))
    return gl, yl, xl



import tensorflow.keras.backend as K
import tensorflow as tf


class Callback(tf.keras.callbacks.Callback):

    def on_epoch_begin(self, epoch, logs=None):
        iterations = K.get_value(self.model.optimizer.iterations)
        self.model.optimizer.start = iterations

    def on_epoch_end(self, epoch, logs=None):
        vs = self.model.trainable_weights
        for v in vs:
            o = self.model.optimizer.get_slot(v, 'old')
            K.set_value(v, K.get_value(o))



import os

import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np

from .model import BaseModel
from .summary import normalized_and_sort
from .summary import summary_stat, summary_footprint_sample, summary_footprint_max


class FootprintModel(BaseModel):

    def __init__(self, footprint, spike, variance, **kwargs):
        super().__init__(footprint, spike, variance, **kwargs)
        self.footprint = footprint
        self.get_spike = spike.get_val
        self.set_spike = spike.set_val
        self.spike_tensor = spike.call

    def call(self, inputs):
        footprint = self.footprint(inputs)
        loss, footprint_penalty, spike_penalty = self.call_common(footprint)
        self.add_metric(footprint_penalty, 'penalty')
        return loss

    def fit(self, spike, lr, batch, verbose, **kwargs):
        nk = spike.shape[0]
        nx = self.variance.nx
        self.set_spike(spike)
        self.start(batch, verbose)
        self.footprint.val = np.zeros((nk, nx))
        self.optimizer.learning_rate = lr * 2.0 / self.variance.lipschitz_a
        self.fit_common(FootprintCallback, verbose=verbose, **kwargs)

    def start(self, batch, verbose):
        data = self.variance._data.enumerate().batch(batch)
        spike = self.spike_tensor()
        calcium = self.variance.spike_to_calcium(spike)
        vdat = K.constant(0.0)
        prog = tf.keras.utils.Progbar(self.variance.nt, verbose=verbose)
        for t, d in data:
            c_p = tf.gather(calcium, t, axis=1)
            vdat += tf.matmul(c_p, d)
            prog.add(tf.size(t).numpy())
        self.variance._cache(1, calcium, vdat)


class FootprintCallback(tf.keras.callbacks.TensorBoard):

    def __init__(self, stage, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage = stage

    def set_model(self, model):
        super().set_model(model)
        self._train_dir = os.path.join(self._log_write_dir, 'footprint')

    def on_epoch_end(self, epoch, logs=None):
        stage = self.stage
        with self._train_writer.as_default():
            val = self.model.footprint.val
            summary_stat(val, stage, step=epoch)
        super().on_epoch_end(logs)

    def on_train_end(self, logs=None):
        stage = self.stage
        with self._train_writer.as_default():
            val = self.model.footprint.val
            mask = self.model.footprint.mask
            summary_footprint_max(val, mask, stage, step=0)
        super().on_train_end(logs)



import tensorflow as tf
import numpy as np


def reduce_peak(peaks, radius, thr_distance):
    idx = reduce_peak_idx(peaks, radius, thr_distance)
    return tuple(v[idx] for v in peaks)


def reduce_peak_idx(peaks, radius, thr_distance):
    ts, rs, ys, xs = peaks[:, 0], peaks[:, 1], peaks[:, 2], peaks[:, 3]
    rs = np.array(radius)[rs]
    total = ts.size
    flg = np.arange(total, dtype=np.int32)
    idx = []
    while flg.size > 0:
        i, j = flg[0], flg[1:]
        y0, x0 = ys[i], xs[i]
        y1, x1 = ys[j], xs[j]
        ya, xa = ys[idx], xs[idx]
        thr = np.square(thr_distance * rs[i])
        if not idx or np.all(np.square(ya - y0) + np.square(xa - x0) >= thr):
            cond = np.square(y1 - y0) + np.square(x1 - x0) >= thr
            flg = j[cond]
            idx.append(i)
        else:
            flg = j
    return np.array(idx, np.int32)



import os

import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..footprint.clean import clean_footprint
from ..image.filter.gaussian import gaussian
from ..train.summary import normalized_and_sort
from ..train.summary import summary_segment
from ..util.numpy import load_numpy, save_numpy


class CleanCommand(Command):

    name = 'clean'
    description = 'Clean segment'
    help = '''The clean command 
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f', ''),
    ]

    def is_error(self, stage):
        return stage < 2

    def is_target(self, stage):
        return stage % 3 == 2

    def force_stage(self, stage):
        return 3 * ((stage - 2) // 3) + 2

    def create(self, data, prev, curr, logs, gauss, radius,
               thr_firmness, thr_sim_area, thr_similarity):
        footprint = load_numpy(f'{prev}-footprint')
        mask = load_numpy(f'{data}-mask')
        batch = self.status.params['batch']
        verbose = self.status.params['pbar']
        thr_out = self.status.params['thr-out']
        radius = np.array(radius)
        nr = radius.size
        segment, pos, firmness = clean_footprint(
            footprint, mask, gauss, radius, batch, verbose,
        )

        idx = np.argsort(firmness)[::-1]
        segment = segment[idx]
        pos = pos[idx]
        fir = firmness[idx]
        rad = radius[pos[:, 0]]

        old_nk = segment.shape[0]
        if thr_sim_area > 0.0:
            h, w = mask.shape
            seg = np.zeros((old_nk, h, w), np.float32)
            seg[:, mask] = segment
            seg = gaussian(seg, gauss).numpy()
            seg -= seg.min(axis=(1, 2), keepdims=True)
            mag = seg.max(axis=(1, 2))
            cond = mag > 0.0
            seg[cond] /= mag[cond, None, None]
            seg[~cond] = 1.0
            seg = seg > thr_sim_area
            cor = np.zeros((old_nk,))
            for j in np.arange(old_nk)[::-1]:
                cj = 0.0
                for i in np.arange(j)[::-1]:
                    ni = np.count_nonzero(seg[i])
                    nij = np.count_nonzero(seg[i] & seg[j])
                    cij = nij / ni
                    if cij > cj:
                        cj = cij
                cor[j] = cj
        else:
            scale = np.sqrt((segment ** 2).sum(axis=1))
            seg = segment / scale[:, None]
            cor = np.zeros((old_nk,))
            for j in np.arange(old_nk)[::-1]:
                cj = 0.0
                for i in np.arange(j)[::-1]:
                    cij = np.dot(seg[i], seg[j])
                    if cij > cj:
                        cj = cij
                cor[j] = cj

        logs = os.path.join(logs, 'clean')
        writer = tf.summary.create_file_writer(logs)
        with writer.as_default():
            fsum = segment.sum(axis=1)
            tf.summary.histogram(f'seg_radius/{curr[-3:]}', rad, step=0)
            tf.summary.histogram(f'seg_firmness/{curr[-3:]}', fir, step=0)
            tf.summary.histogram(f'seg_correlation/{curr[-3:]}', cor, step=0)
            tf.summary.histogram(f'seg_size/{curr[-3:]}', fsum, step=0)
            writer.flush()

            cond = fir > thr_firmness
            cond &= cor < thr_similarity
            cond &= (0 < pos[:, 0]) & (pos[:, 0] < nr - 1)
            idx = np.where(~cond)[0]
            for i in idx:
                print(i, fir[i], cor[i], pos[i, 0])
            summary_segment(segment, mask, cond, gauss, thr_out, curr[-3:])
            writer.flush()

            segment = segment[cond]
            pos = pos[cond]
            rad = rad[cond]
            fir = fir[cond]
            cor = cor[cond]

            fsum = segment.sum(axis=1)
            tf.summary.histogram(f'seg_radius/{curr[-3:]}', rad, step=1)
            tf.summary.histogram(f'seg_firmness/{curr[-3:]}', fir, step=1)
            tf.summary.histogram(f'seg_correlation/{curr[-3:]}', cor, step=1)
            tf.summary.histogram(f'seg_size/{curr[-3:]}', fsum, step=1)
            writer.flush()
        writer.close()

        nk = segment.shape[0]
        self.line(f'ncell: {old_nk} -> {nk}', 'comment')

        save_numpy(f'{curr}-segment', segment)
        save_numpy(f'{curr}-pos', pos[:, 1:])
        save_numpy(f'{curr}-radius', rad)
        save_numpy(f'{curr}-firmness', fir)



import pickle

from .base import Command, _option


class HistoryCommand(Command):

    name = 'history'
    description = 'Show History'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
    ]

    def handle(self):
        self.set_job_dir()
        history = self.status.history
        for k, hs in history.items():
            self.line(f'name: {k}')
            for s, h in enumerate(hs):
                if s == 0:
                    self.line(f'data ({s}): {h[0]}, {h[1]}')
                elif s == 1:
                    self.line(f'peak ({s}): gamma={h[0]}, thr-intensity{h[2]}, shard={h[3]}')
                    self.line('    radius=(' + ', '.join(f'{r:.2f}' for r in h[1]) + ')')
                elif s == 2:
                    self.line(f'segment ({s}): thr-distance{h[0]}')
                elif s % 3 == 0:
                    self.line(f'spike ({s}): tau={h[0]}, lu={h[1]}, bx={h[2]}, bt={h[3]}')
                elif s % 3 == 1:
                    self.line(f'footprint ({s}): la={h[0]}, bx={h[1]}, bt={h[2]}')
                elif s % 3 == 2:
                    self.line(f'{h}')
                    self.line(f'clean ({s}): gamma={h[0]}, thr-firmness={h[2]}, thr-sim-area={h[3]}, thr-similarity={h[4]}')
                    self.line('    radius=(' + ', '.join(f'{r:.2f}' for r in h[1]) + ')')



import tempfile
import os
from urllib.parse import urlparse, urlunparse

import tensorflow as tf


def abspath(filename):
    scheme, netloc, path, params, query, fragment = urlparse(filename)
    if scheme == '':
        scheme = 'file'
        path = os.path.abspath(path)
    return urlunparse((scheme, netloc, path, params, query, fragment))


def ensure_local_file(filename):
    if not filename.startswith('gs://'):
        return filename
    dirname = os.path.dirname(filename[6:])
    dirname = os.path.join(tempfile.gettempdir(), 'hotaru-gs', dirname)
    tf.io.gfile.makedirs(dirname)
    basename = os.path.basename(filename)
    newfilename = os.path.join(dirname, basename)
    tf.io.gfile.copy(filename, newfilename)
    return newfilename



from cleo import Application as ApplicationBase
import tensorflow as tf

from ..version import __version__
from .config import ConfigCommand
from .run import RunCommand
from .history import HistoryCommand
from .data import DataCommand
from .peak import PeakCommand
from .segment import SegmentCommand
from .spike import SpikeCommand
from .footprint import FootprintCommand
from .clean import CleanCommand
from .output import OutputCommand
from .test import TestCommand


class Application(ApplicationBase):

    def __init__(self):
        super().__init__('hotaru', __version__)

        self.job_dir = None
        self.strategy = tf.distribute.MirroredStrategy()

        self.add_commands(
            ConfigCommand(),
            RunCommand(),
            HistoryCommand(),
            DataCommand(),
            PeakCommand(),
            SegmentCommand(),
            SpikeCommand(),
            FootprintCommand(),
            CleanCommand(),
            OutputCommand(),
            TestCommand(),
        )



import os

import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np

from .model import BaseModel
from .summary import summary_stat, summary_spike


class SpikeModel(BaseModel):

    def __init__(self, footprint, spike, variance, **kwargs):
        super().__init__(footprint, spike, variance, **kwargs)
        self.spike = spike
        self.get_footprint = footprint.get_val
        self.set_footprint = footprint.set_val
        self.footprint_tensor = footprint.call

    def call(self, inputs):
        spike = self.spike(inputs)
        calcium = self.variance.spike_to_calcium(spike)
        loss, footprint_penalty, spike_penalty = self.call_common(calcium)
        self.add_metric(footprint_penalty, 'penalty')
        return loss

    def fit(self, footprint, lr, batch, verbose, **kwargs):
        nk = footprint.shape[0]
        nu = self.variance.nu
        self.set_footprint(footprint)
        self.start(batch, verbose)
        self.spike.val = np.zeros((nk, nu))
        self.optimizer.learning_rate = lr * 2.0 / self.variance.lipschitz_u
        self.fit_common(SpikeCallback, verbose=verbose, **kwargs)

    def start(self, batch, verbose):
        data = self.variance._data.batch(batch)
        footprint = self.footprint_tensor()
        adat = tf.TensorArray(tf.float32, 0, True)
        prog = tf.keras.utils.Progbar(self.variance.nt, verbose=verbose)
        for d in data:
            adat_p = tf.matmul(d, footprint, False, True)
            for p in adat_p:
                adat = adat.write(adat.size(), p)
                prog.add(1)
        self.variance._cache(0, footprint, tf.transpose(adat.stack()))


class SpikeCallback(tf.keras.callbacks.TensorBoard):

    def __init__(self, stage, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage = stage

    def set_model(self, model):
        super().set_model(model)
        self._train_dir = os.path.join(self._log_write_dir, 'spike')

    def on_epoch_end(self, epoch, logs=None):
        stage = self.stage
        with self._train_writer.as_default():
            summary_stat(self.model.spike.val, stage, step=epoch)
        super().on_epoch_end(epoch, logs)

    def on_train_end(self, logs=None):
        stage = self.stage
        with self._train_writer.as_default():
            summary_spike(self.model.spike.val, stage, step=0)
        super().on_train_end(logs)



import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np
from scipy.ndimage import gaussian_filter as gaussian
from scipy.ndimage import label, binary_closing
from scipy.ndimage import generate_binary_structure, binary_dilation
from matplotlib import cm


_jet = cm.get_cmap('jet')
_reds = cm.get_cmap('Reds')
_greens = cm.get_cmap('Greens')


def normalized_and_sort(val):
    mag = val.max(axis=1)
    idx = np.argsort(mag)[::-1]
    val = val / mag[:, None]
    val = val[idx]
    return val, mag


def summary_stat(val, stage, step=0):
    val, mag = normalized_and_sort(val)
    spc = val.mean(axis=1)
    tf.summary.histogram(f'val_max/{stage}', mag, step=step)
    tf.summary.histogram(f'val_avg/{stage}', spc, step=step)


def summary_spike(val, stage, step=0):
    val, mag = normalized_and_sort(val)
    tf.summary.image(f'spike/{stage}', _reds(val)[None, ...], step=step)


def summary_footprint_sample(val, mask, stage, step=0):
    h, w = mask.shape
    imgs = np.zeros((4, h, w))
    imgs[:, mask] = val[[0, 1, -2, -1]]
    tf.summary.image(
        f'footprint-sample/{stage}',
        _greens(imgs), max_outputs=4, step=step,
    )


def summary_footprint_max(val, mask, stage, step=0):
    val_max = val.max(axis=0)
    val_max /= val.max()
    h, w = mask.shape
    imgs_max = np.zeros((1, h, w))
    imgs_max[0, mask] = val_max
    tf.summary.image(f'max/{stage}', _greens(imgs_max), step=step)


def summary_segment(val, mask, flag, gauss, thr_out, stage):
    struct = generate_binary_structure(2, 2)
    h, w = mask.shape
    out = 0
    b0 = False
    b1 = False
    for f, v in zip(flag[::-1], val[::-1]):
        img = np.zeros((h, w))
        img[mask] = v
        g = gaussian(img, gauss)
        g /= g.max()
        peak = np.argmax(g)
        lbl, n = label(g > thr_out)
        size = 0
        for i in range(1, n+1):
            tmp = binary_closing(lbl == i)
            ts = np.count_nonzero(tmp)
            if ts > size:
                size = ts
                obj = tmp
        out += obj
        bound = binary_dilation(obj, struct) ^ obj
        if f:
            b0 |= bound
        else:
            b1 |= bound
    out = out / out.max()
    out = _greens(out)
    out[b0] = (0, 0, 0, 1)
    out[b1] = (1, 0, 0, 1)
    tf.summary.image(f'segment/{stage}', out[None, ...], step=0)



import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..footprint.reduce import reduce_peak_idx
from ..footprint.make import make_segment
from ..train.summary import summary_segment
from ..util.tfrecord import load_tfrecord
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle


class SegmentCommand(Command):

    name = 'segment'
    description = 'Make segment'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 2

    def is_target(self, stage):
        return stage == 2

    def force_stage(self, stage):
        return 2

    def create(self, data, prev, curr, logs, thr_intensity, thr_distance):
        batch = self.status.params['batch']
        verbose = self.status.params['pbar']
        thr_out = self.status.params['thr-out']

        tfrecord = load_tfrecord(f'{data}-data')
        mask = load_numpy(f'{data}-mask')
        gauss, radius, shard = load_pickle(f'{prev}-filter')
        radius = np.array(radius)
        nr = radius.size
        p = load_numpy(f'{prev}-peak')
        r = radius[p[:, 1]]
        s = load_numpy(f'{prev}-intensity')

        writer = tf.summary.create_file_writer(logs)
        with writer.as_default():
            cond = s > thr_intensity
            p = p[cond]
            r = r[cond]
            s = s[cond]
            tf.summary.histogram(f'init/radius', r, step=0)
            tf.summary.histogram(f'init/intensity', s, step=0)
            writer.flush()

            idx = reduce_peak_idx(p, radius, thr_distance)
            p = p[idx]
            r = r[idx]
            s = s[idx]
            tf.summary.histogram(f'init/radius', r, step=1)
            tf.summary.histogram(f'init/intensity', s, step=1)
            writer.flush()

            cond = (0 < p[:, 1]) & (p[:, 1] < nr - 1)
            p = p[cond]
            r = r[cond]
            s = s[cond]
            tf.summary.histogram(f'init/radius', r, step=2)
            tf.summary.histogram(f'init/intensity', s, step=2)
            writer.flush()

            segment = make_segment(
                tfrecord, mask, gauss, radius, p, shard, batch, verbose,
            )

            nk = segment.shape[0]
            cond = np.ones(nk, np.bool)
            summary_segment(segment, mask, cond, gauss, thr_out, curr[-3:])
            fsum = segment.sum(axis=1)
            tf.summary.histogram(f'init/seg_size', fsum, step=0)
            writer.flush()
        writer.close()

        save_numpy(f'{curr}-segment', segment)
        save_numpy(f'{curr}-peak', p[:, [0, 2, 3]])
        save_numpy(f'{curr}-radius', r)
        save_numpy(f'{curr}-intensity', s)



import tensorflow as tf
import numpy as np


def load_csv(base):
    with tf.io.gfile.GFile(f'{base}.csv', 'r') as fp:
        return np.loadtxt(fp, delimiter=',', skiprows=1)


def save_csv(base, data, fmt, header):
    with tf.io.gfile.GFile(f'{base}.csv', 'w') as fp:
        np.savetxt(fp, data, fmt, ',', header=header)



import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..train.spike import SpikeModel
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle, save_pickle


class SpikeCommand(Command):

    description = 'Update spike'

    name = 'spike'
    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 3

    def is_target(self, stage):
        return stage % 3 == 0

    def force_stage(self, stage):
        return 3 * ((stage - 0) // 3) + 0

    def create(self, data, prev, curr, logs, tau, lu, bx, bt):
        segment = load_numpy(f'{prev}-segment')
        nk = segment.shape[0]
        la = self.status.params['la']
        #with self.application.strategy.scope():
        elems = self.get_model(data, tau, nk)
        model = SpikeModel(*elems)
        model.compile()
        model.set_penalty(la, lu, bx, bt)
        model.fit(
            segment, stage=curr[-3:],
            lr=self.status.params['learning-rate'],
            batch=self.status.params['batch'],
            steps_per_epoch=self.status.params['step'],
            epochs=self.status.params['epoch'],
            verbose=self.status.params['pbar'],
            min_delta=self.status.params['tol'],
            log_dir=logs,
        )
        save_pickle(f'{curr}-tau', tuple(tau))
        save_numpy(f'{curr}-spike', model.spike.val)



import tensorflow as tf


def save_tfrecord(filebase, data, nt=None, verbose=1):
    prog = tf.keras.utils.Progbar(nt, verbose=verbose)
    with tf.io.TFRecordWriter(f'{filebase}.tfrecord') as writer:
        for d in data:
            writer.write(tf.io.serialize_tensor(d).numpy())
            if prog is not None:
                prog.add(1)


def load_tfrecord(filebase):
    data = tf.data.TFRecordDataset(f'{filebase}.tfrecord')
    data = data.map(lambda ex: tf.io.parse_tensor(ex, tf.float32))
    return data



import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np


class SpikeToCalcium(tf.keras.layers.Layer):

    def set_double_exp(self, hz, tau1, tau2, tscale):
        if tau1 < tau2:
            tau1, tau2 = tau1, tau2
        if hz is not None:
            tau1 *= hz
            tau2 *= hz
        r = tau1 / tau2
        d = tau1 - tau2
        scale = np.power(r, -tau2 / d) - np.power(r, -tau1 / d)
        t = np.arange(1.0, tau1 * tscale + 2.0, dtype=np.float32)
        e1 = np.exp(-t / tau1)
        e2 = np.exp(-t / tau2)
        kernel = (e1 - e2) / scale
        size = kernel.size
        if hasattr(self, 'kernel'):
            if kernel.size > K.int_shape(self.kernel)[0]:
                raise RuntimeError(f'filter size mismatch: {size}')
        else:
            self.kernel = self.add_weight('kernel', (size,), trainable=False)
            self.pad = size - 1
        K.set_value(self.kernel, kernel)

    def call(self, u):
        return K.conv1d(
            u[..., None], self.kernel[::-1, None, None],
            1, 'valid', 'channels_last',
        )[..., 0]


class CalciumToSpike(tf.keras.layers.Layer):

    def set_double_exp(self, hz, tau1, tau2, tscale):
        if tau1 < tau2:
            tau1, tau2 = tau1, tau2
        if hz is not None:
            tau1 *= hz
            tau2 *= hz
        r = tau1 / tau2
        d = tau1 - tau2
        scale = np.power(r, -tau2 / d) - np.power(r, -tau1 / d)
        e1 = np.exp(-1 / tau1)
        e2 = np.exp(-1 / tau2)
        kernel = np.array([1.0, -e1 - e2, e1 * e2]) / (e1 - e2) * scale
        size = kernel.size
        if hasattr(self, 'kernel'):
            if size > K.int_shape(self.kernel)[0]:
                raise RuntimeError(f'filter size mismatch: {size}')
        else:
            self.kernel = self.add_weight('kernel', (size,), trainable=False)
            self.pad = int(tau1 * tscale + 1)
        K.set_value(self.kernel, kernel)

    def call(self, v):
        v = tf.pad(v, [[0, 0], [2, 0]], 'CONSTANT')
        u = K.conv1d(
            v[..., None], self.kernel[::-1, None, None],
            1, 'valid', 'channels_last',
        )[..., 0]
        return tf.pad(u, [[0, 0], [self.pad, 0]], 'CONSTANT')



import tensorflow as tf

from .regularizer import get_prox


class ProxNesterov(tf.keras.optimizers.Optimizer):

    _HAS_AGGREGATE_GRAD = True

    def __init__(self, learning_rate=1.0, scale=20.0,
                 name='ProxNesterov', **kwargs):
        super().__init__(name, **kwargs)
        self._set_hyper('learning_rate', kwargs.get('lr', learning_rate))
        self._set_hyper('scale', scale)
        self._set_hyper('start', 0)

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, 'old')

    def _prepare_local(self, var_device, var_dtype, apply_state):
        super()._prepare_local(var_device, var_dtype, apply_state)
        scale = self._get_hyper('scale', var_dtype)
        step = tf.cast(self.iterations - self.start, var_dtype)
        apply_state[(var_device, var_dtype)].update(dict(
            t0=(scale + step) / scale,
            t1=(scale + 1) / (scale + step + 1),
        ))

    def _resource_apply_dense(self, grad, var, apply_state=None):
        key = var.device, var.dtype.base_dtype
        coef = apply_state.get(key)
        lr = coef['lr_t']
        t0 = coef['t0']
        t1 = coef['t1']
        old = self.get_slot(var, 'old')
        new_t = get_prox(var)(var - lr * grad, lr)
        tmp_t = (1.0 - t0) * old + t0 * new_t
        upd_t = (1.0 - t1) * new_t + t1 * tmp_t
        return tf.group(
            old.assign(new_t, use_locking=self._use_locking),
            var.assign(upd_t, use_locking=self._use_locking),
        )

    def get_config(self):
        config = super().get_config()
        config.update(dict(
            learning_rate=self._serialize_hyperparameter('learning_rate'),
            scale=self._serialize_hyperparameter('scale'),
        ))
        return config



import tensorflow.keras.backend as K
import tensorflow as tf

from ..util.distribute import distributed, ReduceOp
from ..util.dataset import unmasked
from ..image.filter.gaussian import gaussian
from ..image.filter.laplace import gaussian_laplace_multi
from .segment import get_segment_index
from .util import get_normalized_val, get_magnitude, ToDense


def clean_footprint(data, mask, gauss, radius, batch, verbose):
    strategy = tf.distribute.get_strategy()

    dataset = tf.data.Dataset.from_tensor_slices(data)
    dataset = unmasked(dataset, mask)
    dataset = dataset.batch(batch)
    to_dense = ToDense(mask)

    mask = K.constant(mask, tf.bool)
    gauss = K.constant(gauss)
    radius = K.constant(radius)

    nk = data.shape[0]
    prog = tf.keras.utils.Progbar(nk, verbose=verbose)
    ss = tf.TensorArray(tf.float32, nk)
    ps = tf.TensorArray(tf.int32, nk)
    fs = tf.TensorArray(tf.float32, nk)
    i = tf.constant(0)
    for data in strategy.experimental_distribute_dataset(dataset):
        gl, ll, rl, yl, xl = _prepare(data, mask, gauss, radius)
        nx = tf.size(rl)
        for k in tf.range(nx):
            g, l, r, y, x = gl[k], ll[k], rl[k], yl[k], xl[k]
            pos = get_segment_index(l, y, x, mask)
            val = get_normalized_val(g, pos)
            footprint = to_dense(pos, val)
            firmness = get_magnitude(l, pos) / get_magnitude(g, pos)
            ss = ss.write(i, footprint)
            ps = ps.write(i, [r, y, x])
            fs = fs.write(i, firmness)
            i += 1
            prog.add(1)
    return ss.stack().numpy(), ps.stack().numpy(), fs.stack().numpy()


@distributed(*[ReduceOp.CONCAT for _ in range(5)], loop=False)
def _prepare(imgs, mask, gauss, radius):
    gs = gaussian(imgs, gauss) if gauss > 0.0 else imgs
    ls = gaussian_laplace_multi(gs, radius)
    nk, h, w = tf.shape(ls)[0], tf.shape(ls)[2], tf.shape(ls)[3]
    hw = h * w
    lsr = K.reshape(ls, (nk, -1))
    pos = tf.cast(K.argmax(lsr, axis=1), tf.int32)
    rs = pos // hw
    ys = (pos % hw) // w
    xs = (pos % hw) % w
    ls = tf.gather_nd(ls, tf.stack([tf.range(nk), rs], axis=1))
    return imgs, ls, rs, ys, xs



import numpy as np


class Status(object):

    def __init__(self):
        self.params = dict()
        self.history = dict()
        self.saved = dict()

    def add_saved(self, key, name):
        self.saved[key] = name

    def find_saved(self, key):
        return self.saved.get(key, None)

    def get_params(self, stage):
        p = self.params
        if stage == 0:
            return p['imgs-file'], p['mask-type'],
        if stage == 1:
            return p['gauss'], self.radius, p['min-intensity'], p['min-distance'], p['shard'],
        if stage == 2:
            return p['thr-intensity'], p['thr-distance'],
        if stage % 3 == 0:
            return self.tau, p['lu'], p['bx'], p['bt'],
        if stage % 3 == 1:
            return p['la'], p['bx'], p['bt'],
        if stage % 3 == 2:
            return p['gauss'], self.radius, p['thr-firmness'], p['thr-sim-area'], p['thr-similarity'],

    def get_test_params(self):
        p = self.params
        return p['min-intensity'], p['max-intensity'], p['min-distance'], p['max-distance'],

    @property
    def tau(self):
        p = self.params
        return p['hz'], p['tau-rise'], p['tau-fall'], p['tau-scale'],

    @property
    def radius(self):
        radius_type = self.params['radius-type']
        r = self.params['radius']
        if radius_type == 'linear':
            out = np.linspace(r[0], r[1], int(r[2]))
        elif radius_type == 'log':
            out = np.logspace(np.log10(r[0]), np.log10(r[1]), int(r[2]))
        elif radius_type == 'manual':
            out = r
        else:
            raise RuntimeError(f'invalid radius type: {radius_type}')
        return tuple(out)



import datetime
import os

import tensorflow as tf
import numpy as np
import matplotlib.cm as cm

from .base import Command, option, _option
from ..footprint.reduce import reduce_peak_idx
from ..util.numpy import load_numpy
from ..util.pickle import load_pickle


class TestCommand(Command):

    name = 'test'
    description = 'Test to make segment'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
    ]

    def handle(self):
        self.set_job_dir()

        name = self.status.params['name']

        stage = len(self.status.history.get(name, ()))
        if stage <= 0:
            self.call('data')
        if stage <= 1:
            self.call('peak')

        self.line('test', 'info')
        min_intensity, max_intensity, min_distance, max_distance = self.status.get_test_params()
        green = cm.Greens

        history = self.status.history.get(name, ())
        data = self.status.find_saved(history[:1])
        data = os.path.join(self.work_dir, data, f'000')
        h, w = load_numpy(f'{data}-mask').shape
        prev = self.status.find_saved(history[:2])
        prev = os.path.join(self.work_dir, prev, f'001')
        radius = np.array(load_pickle(f'{prev}-filter')[1])
        nr = radius.size
        p0 = load_numpy(f'{prev}-peak')
        r0 = radius[p0[:, 1]]
        s0 = load_numpy(f'{prev}-intensity')

        dt = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
        logs = os.path.join(self.application.job_dir, 'logs', name, f'test-{dt}')
        writer = tf.summary.create_file_writer(logs)
        with writer.as_default():
            for step, thr_distance in enumerate(np.linspace(min_distance, max_distance, 12)):
                idx = reduce_peak_idx(p0, radius, thr_distance)
                p1 = p0[idx]
                r1 = r0[idx]
                s1 = s0[idx]
                tf.summary.histogram(f'test/radius', r1, step=step)
                tf.summary.histogram(f'test/intensity', s1, step=step)
                writer.flush()
                for thr_intensity in np.linspace(min_intensity, max_intensity, 10):
                    cond = s1 > thr_intensity
                    cond &= (0 < p1[:, 1]) & (p1[:, 1] < nr - 1)
                    img = np.zeros((h, w), np.int32)
                    xx, yy = np.meshgrid(np.arange(w), np.arange(h))
                    for (y, x), r in zip(p1[cond, 2:], r1[cond]):
                        cond = (xx - x) ** 2 + (yy - y) ** 2 < (1.5 * r) ** 2
                        img[cond] += 1
                    img = green(img / img.max())
                    tf.summary.image(f'test/{thr_intensity:.3f}-{thr_distance:.3f}', img[None, ...], step=0)
                    writer.flush()
        writer.close()



import tensorflow as tf
import numpy as np
import tifffile

from ..util.gs import ensure_local_file


def get_shape(filename):
    filename = ensure_local_file(filename)
    in_type = filename.split('.')[-1]
    if in_type == 'npy':
        return np.load(filename, mmap_mode='r').shape
    elif in_type == 'tif':
        return tifffile.TiffFile(filename).series[0].shape
    elif in_type[:3] == 'raw':
        _, dtype, h, w, endian = in_type.split('_')
        h, w = int(h), int(w)
        dtype = np.dtype(dtype).newbyteorder('<' if endian == 'l' else '>')
        return np.memmap(filename, dtype, 'r', shape=(-1, h, w)).shape
    raise RuntimeError(f'{filename} is not imgs file')


def load_data(filename):
    filename = ensure_local_file(filename)
    in_type = filename.split('.')[-1]
    if in_type == 'npy':
        return load_data_npy(filename)
    elif in_type == 'tif':
        return load_data_tif(filename)
    elif in_type[:3] == 'raw':
        return load_data_raw(filename)
    raise RuntimeError(f'{filename} is not imgs file')


def load_data_npy(filename):
    mmap = np.load(filename, mmap_mode='r')
    return mmap, lambda x: x


def load_data_tif(filename):
    tif = tifffile.TiffFile(filename)
    if tif.series[0].offset:
        mmap = tif.series[0].asarray(out='memmap')
        return mmap, lambda x: x
    else:
        return tif.series[0], lambda x: x.asarray()


def load_data_raw(filename):
    in_type = infile.split('.')[-1]
    _, dtype, h, w, endian = in_type.split('_')
    h, w = int(h), int(w)
    dtype = np.dtype(dtype).newbyteorder('<' if endian == 'l' else '>')
    mmap = np.memmap(filename, dtype, 'r', shape=(-1, h, w))
    return mmap, lambda x: x



import tensorflow as tf
import numpy as np


def save_numpy(base, val):
    with tf.io.gfile.GFile(f'{base}.npy', 'bw') as fp:
        np.save(fp, val)


def load_numpy(base):
    with tf.io.gfile.GFile(f'{base}.npy', 'br') as fp:
        return np.load(fp)



import tempfile

import tensorflow as tf
import tifffile


def load_tiff(base):
    with tf.io.gfile.GFile(f'{base}.tif', 'rb') as fp:
        return tifffile.imload(fp)


def save_tiff(base, val):
    if base.startswith('gs://'):
        dirname = os.path.dirname(base[6:])
        dirname = os.path.join(tempfile.gettempdir(), 'hotaru-gs', dirname)
        tf.io.gfile.makedirs(dirname)
        basename = os.path.basename(base)
        localfile = os.path.join(dirname, f'{basename}.tif')
        tifffile.imwrite(localfile, val)
        tf.io.gfile.copy(localfile, f'{base}.tif')
    else:
        tifffile.imwrite(f'{base}.tif', val)



import tensorflow.keras.backend as K

from ..util.distribute import distributed, ReduceOp
from .filter.neighbor import neighbor


def calc_cor(imgs):

    @distributed(*[ReduceOp.SUM for _ in range(6)])
    def _calc(img):
        nei = neighbor(img)
        sx1 = K.sum(img, axis=0)
        sy1 = K.sum(nei, axis=0)
        sx2 = K.sum(K.square(img), axis=0)
        sxy = K.sum(img * nei, axis=0)
        sy2 = K.sum(K.square(nei), axis=0)
        ntf = K.cast_to_floatx(K.shape(img)[0])
        return xs1, sy1, sx2, xsy, sy2, ntf

    xs1, xy1, xs2, xsy, xy2, ntf = _calc(imgs)
    avg_x = sx1 / ntf
    avg_y = sy1 / ntf
    cov_xx = sx2 / ntf - K.square(avg_x)
    cov_xy = sxy / ntf - avg_x * avg_y
    cov_yy = sy2 / ntf - K.square(avg_y)
    return cov_xy / K.sqrt(cov_xx * cov_yy)
